//     ======================= RecognitionApp.cpp ===========================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 04.01.2010
//     **********************************************************************

#pragma warning (disable:4996) // This function or variable may be unsafe

#include <stdio.h>

#include "wx/wxprec.h"
#include "RecognitionApp.h"
#include "Russification.h"

#include <wx/colordlg.h>
#include <wx/checkbox.h>
#include <wx/file.h>
#include <wx/image.h>
#include <wx/busyinfo.h>
#include <wx/propdlg.h>
#include <wx/filedlg.h>
#include <wx/spinctrl.h>
#include <wx/menu.h>
#include "wx/generic/filedlgg.h"
#include "wx/stdpaths.h"

#define cStrLearnObj _MST("Learn object") //  " "

CRecognitionPipeline* gpRcgPipeline = NULL; //    
double         gPrfTime = 0;                //      
double         gEOT = 0;                    //    
bool           gShowInputImage = false;     //  "  "
bool           gStopWriteDemo = true;       //  "  "
CvVideoWriter* gOutAVI = NULL;              //     
IplImage*      gOutImg = NULL;              //      

//=============================================================================================
//================================= Class: CDefineObjectName ==================================
//=============================================================================================

//    "  "
class CDefineObjectName : public wxDialog
{
	wxComboBox* mpComboBox; //  
	TObjName*   pObjName;   //   
public:
	//  
    CDefineObjectName(TObjName* apObjName);
	//   "Ok"
    void OnOkButton(wxCommandEvent& event);

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(CDefineObjectName, wxDialog)
	EVT_BUTTON(wxID_OK, CDefineObjectName::OnOkButton)
END_EVENT_TABLE()

//  
CDefineObjectName::CDefineObjectName(TObjName* apObjName) :
              wxDialog(NULL, wxID_ANY, _MST("Enter object name"),
              wxDefaultPosition, wxSize(300,100), wxDEFAULT_DIALOG_STYLE|wxMINIMIZE_BOX)
{
	pObjName = apObjName;
	mpComboBox = new wxComboBox(this, -1, _T("New object"));
	for(int i=0; i<pObjName->Total; i++) {
        mpComboBox->Append(wxString::FromAscii(pObjName->s[i]));
    }
	wxBoxSizer *mpTopSizer = new wxBoxSizer(wxVERTICAL);
    mpTopSizer->Add(mpComboBox, 1, wxGROW | wxALL, 5);
	wxBoxSizer *ButtonSizer = new wxBoxSizer(wxHORIZONTAL);
	ButtonSizer->Add(new wxButton(this, wxID_OK),     0, wxALL, 5);
	ButtonSizer->Add(new wxButton(this, wxID_CANCEL), 0, wxALL, 5);	
	mpTopSizer->Add(ButtonSizer, 0, wxALL | wxALIGN_CENTRE_HORIZONTAL, 5);

	SetSizer(mpTopSizer);
}
			  
//   "Ok"
void CDefineObjectName::OnOkButton(wxCommandEvent& WXUNUSED(event))
{

	pObjName->ObjIdx = mpComboBox->GetSelection();
	wxString ObjName = mpComboBox->GetValue();

	for(int i=0; i<pObjName->Total; i++) {
		if(!ObjName.Cmp(wxString::FromAscii(pObjName->s[i]))) {
			pObjName->ObjIdx = i;
			break;
		}
	}

	if(pObjName->ObjIdx < 0) {
		pObjName->ObjIdx = pObjName->Total;
		strcpy(pObjName->s[pObjName->Total++], ObjName.ToAscii());
	}
	
	EndModal(wxID_OK);
}

// : " "
//=============================================================================================
//=============================== Class: CRecognitionPipeline =================================
//=============================================================================================
//   
struct TGateData {
	int    cX,cY; //   
	CvRect     r; //    
	CvType192U d; // 
};

typedef iList<TGateData> TGateList;

//      "doSearchObject"
struct TSearchObjectData {
	IplImage* OutImg;     //      
	TObjName* pObjName;   //     
	TRcgObj   RcgObj;     //   
	TGateList GateList;   //   ""
};

//   
void doSearchObject(CvAM_State aState, long* apData, CvAM_Parameters* apParam, void* args)
{
	char OutStr[256];
	int i;
	static double bTime,eTime,Tf; //      
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );
	CvFont SmallFont;
	cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );
	
	TSearchObjectData* d = (TSearchObjectData*)args;

	IplImage* pImg = d->OutImg;

	CvRect ObjRect = cvRect( 0, 0, 0, 0 );

	if(apParam) {
		//      Y
		ObjRect   = apParam->ObjRect;
		ObjRect.y = pImg->height - apParam->ObjRect.y;
	}

	switch(aState) {
		case cProcessingIsBegun: //    
			bTime = (double)cvGetTickCount();
			Tf = (double)cvGetTickFrequency()*1000.0;
			break;

		case cRecognizedObject: //  (  )  
			cvRectangle( pImg, cvPoint(ObjRect.x,  ObjRect.y),
							   cvPoint(ObjRect.x + ObjRect.width,
									   ObjRect.y - ObjRect.height),cvScalar(255,128,0),1,8,0);
			break;

		case cGeneralizedObject:
		case cTrackedObject: //  "" 
			if(d->RcgObj.ObjTotal < cObjTotalMax) {
				//     
				d->RcgObj.ObjIdx[d->RcgObj.ObjTotal] = *apData;
				d->RcgObj.ObjRect[d->RcgObj.ObjTotal++] = ObjRect;
			}
			//    ( )
			cvRectangle( pImg, cvPoint(ObjRect.x,  ObjRect.y),
							   cvPoint(ObjRect.x + ObjRect.width,
									   ObjRect.y - ObjRect.height),
									   cvScalar(255,0,0),2,8,0);

			//   
			sprintf(OutStr, "%s", d->pObjName->s[*apData]);
			cvPutText(pImg, OutStr, cvPoint(ObjRect.x+10, ObjRect.y -
											   ObjRect.height - 24), &font, CV_RGB(255,255,0));

			//     ()
			if(apParam->ppUserTrackingInfo) {
				#define cTrcLen 100
				struct TTrcInfo {
					CvPoint TrcPoint[cTrcLen];
					long Age;
				};
				if(!*apParam->ppUserTrackingInfo) {
					*apParam->ppUserTrackingInfo = new TTrcInfo;
					memset(*apParam->ppUserTrackingInfo, 0, sizeof(TTrcInfo));
				} else {
					TTrcInfo* pTrcInfo = (TTrcInfo*)*apParam->ppUserTrackingInfo;
					memcpy(&pTrcInfo->TrcPoint[0], &pTrcInfo->TrcPoint[1], sizeof(CvPoint)*(cTrcLen-1));
					pTrcInfo->TrcPoint[cTrcLen-1] = cvPoint(ObjRect.x + (ObjRect.width>>1),
															ObjRect.y - (ObjRect.height>>1));
					if(pTrcInfo->Age < cTrcLen) { pTrcInfo->Age++; }

					if(pTrcInfo->Age > 2) {
						for(i=1; i<pTrcInfo->Age; i++) {
							//    
							cvLine(pImg, pTrcInfo->TrcPoint[cTrcLen-i-1],
										 pTrcInfo->TrcPoint[cTrcLen-i], cvScalar(255,255,0),2,8,0);
						}
						//   
						cvCircle(pImg, pTrcInfo->TrcPoint[cTrcLen-1],4,cvScalar(0,255,0),2,8,0);
					}
				}
			}
			break;

		case cLearnThisObject: //    (  )
			cvRectangle( pImg, cvPoint(ObjRect.x,  ObjRect.y),
							   cvPoint(ObjRect.x + ObjRect.width,
								       ObjRect.y - ObjRect.height),
									   cvScalar(255,255,0),2,8,0);
			break;

		case cProcessingIsFinished: //    
			eTime = (double)cvGetTickCount(); gPrfTime = (eTime-bTime)/Tf; 
			
			sprintf(OutStr,"%5.1f ms  %lu/%6.4f",gPrfTime, apParam->TotalABases, apParam->Similarity);
			//   ( /  / )
			cvPutText(pImg, OutStr, cvPoint(0, pImg->height - 14), &SmallFont, CV_RGB(128,255,0));
			break;
	}
}

//   
void doSearchGate(CvAM_State aState, CvType192U* apData, CvAM_Parameters* apParam, void* args)
{
	char OutStr[256];
	static double bTime,eTime,Tf; //      
	CvFont SmallFont;
	cvInitFont( &SmallFont, 1, 1.0, 1.0, 0, 1, 8 );
	
	TSearchObjectData* d = (TSearchObjectData*)args;

	IplImage* pImg = d->OutImg;

	CvRect ObjRect = cvRect( 0, 0, 0, 0 );

	if(apParam) {
		//      Y
		ObjRect   = apParam->ObjRect;
		ObjRect.y = pImg->height - apParam->ObjRect.y;
	}

	switch(aState) {
		case cProcessingIsBegun: //    
			bTime = (double)cvGetTickCount();
			Tf = (double)cvGetTickFrequency()*1000.0;
			break;

		case cRecognizedObject: //  (  )  
			cvRectangle( pImg, cvPoint(ObjRect.x,  ObjRect.y),
							   cvPoint(ObjRect.x + ObjRect.width,
									   ObjRect.y - ObjRect.height),cvScalar(128,0,255),1,8,0);
			break;

		case cGeneralizedObject:
		case cTrackedObject: //  "" 
			TGateData GateData;
			GateData.cX = apParam->ObjRect.x + (apParam->ObjRect.width>>1);
			GateData.cY = apParam->ObjRect.y + (apParam->ObjRect.height>>1);
			GateData.r  = ObjRect;
			GateData.d  = *apData;
			d->GateList.Append(GateData);
			
			if(apData->s[cPointIdx_s] == (gWayIdx+1)) { d->RcgObj.CheckpointFound = true; }

			break;

		case cProcessingIsFinished: //    
			eTime = (double)cvGetTickCount(); gPrfTime = (eTime-bTime)/Tf; 
			
			sprintf(OutStr,"%5.1f ms  %lu/%6.4f  Go to p[%d]",
				    gPrfTime, apParam->TotalABases, apParam->Similarity, gWayIdx+1);
			//   ( /  / )
			cvPutText(pImg, OutStr, cvPoint(0, pImg->height - 32), &SmallFont, CV_RGB(255,255,0));
			break;
	}
}


BEGIN_EVENT_TABLE(CRecognitionPipeline, wxStaticText)
	EVT_MOTION(CRecognitionPipeline::OnMouseMove)
	EVT_LEFT_DOWN(CRecognitionPipeline::OnMouseLeftDown)
END_EVENT_TABLE()

//  
CRecognitionPipeline::CRecognitionPipeline(wxWindow* parent, int aCamIdx,
	TCaptureSettings aCaptSet, VidFormat* apVidFmt) : wxStaticText(parent, wxID_ANY,
	_T("Recognition pipeline"), wxDefaultPosition, InitCapture(aCamIdx, aCaptSet, apVidFmt))
{
	mpObjAM  = NULL;
	mpGateAM = NULL;
	mpGray   = NULL;
	mpOptFlw = NULL;
	mpDrawTree = NULL;
	mMouseCursorInit = false;
	mAspectRatio = 1.0;
	mComment = _T("");
	mCommentTimer = 0;
	memset(&mRcgDataFile, 0, sizeof(TRcgDataFile));
	cvInitFont( &mHint_Font, CV_FONT_HERSHEY_PLAIN, 1.0, 1.0, 0, 1, 8 );
	cvInitFont( &mComment_Font, 2, 0.7, 0.7, 0.3, 1, 8 );
}

//  
CRecognitionPipeline::~CRecognitionPipeline() {
	if(mpGray) {
		
		cvcamStop();
		cvcamExit();

		if(mpGateAM) { delete mpGateAM; }
		if(mpObjAM) { delete mpObjAM; }

		cvReleaseImage(&mpGray);	
	}

	if(mpOptFlw) { delete mpOptFlw; }
	if(mpDrawTree) { delete mpDrawTree; }
}


#define cPauseLoopMax    50  //     
#define cTrainingLoopMax 100 //   

//    
void CRecognitionPipeline::SetTheBaseCoordinateOfObject()
{
	mWorkMode = cSetTheBaseCoordinateOfObject;
	if(mNeedKeyImageSetting) {
		mInterestArea.width  = cBaseKeyImgSize;
		mInterestArea.height = cBaseKeyImgSize;
		mKeyImageSize.width  = cBaseKeyImgSize;
		mKeyImageSize.height = cBaseKeyImgSize;
		wxLogMessage(_MST("\nNow you must set a key image size for associative video memory."));
	} else {
		CvSize InterestAreaSize = mpObjAM->GetKeyImageSize();
		mInterestArea.width  = InterestAreaSize.width;
		mInterestArea.height = InterestAreaSize.height;		
	}
	wxLogMessage(_MST("Set the cursor at left top corner of an object and click a left mouse button."));
	mPauseLoop    = cPauseLoopMax;
	mTrainingLoop = cTrainingLoopMax;
	wxSize HintWinSize = wxSize(110,80);
	mMouseCursorX = 0;
	mMouseCursorY = 0;
	SetHintWin(_T("Fix it at\n")
		       _T("the left top\n")
			   _T("corner of\n")
			   _T("an object"), &HintWinSize);
}

//    
void LogMessageFunc(char* apMsg) {
	wxLogMessage(wxString::FromAscii(apMsg));
}

//   
void SetCheckpoint(int aWayNum, ushort aGateCode) {
	gpRcgPipeline->SetCheckpoint(aWayNum, aGateCode);
}

// -  
void RecognitionPipelineFunc(IplImage* apImage)
{
#define pRPL gpRcgPipeline

	char OutStr[256];
	CvFont font;
	cvInitFont( &font, 2, 1.0, 1.0, 0, 2, 8 );

	
	assert(apImage);
	
	pRPL->EnterProcessing();
	pRPL->mpScreen = apImage;
	
	cvCvtColor(apImage, pRPL->mpGray, CV_BGR2GRAY);
	cvFlip(pRPL->mpGray);

	if(gShowInputImage) {
		//   
		IplImage* cGray = cvCreateImage(cvSize(pRPL->mpGray->width,
											   pRPL->mpGray->height),
											   IPL_DEPTH_8U, 1);
		cvCopy(pRPL->mpGray, cGray);
		
		switch(pRPL->mWorkMode) {

			case cSetTheKeyImageSize:
			case cPrepareToTraining:
			case cTraining:
			
				IplImage* ciGray = cvCreateImage(cvSize(pRPL->mInterestArea.width,
														pRPL->mInterestArea.height),
														IPL_DEPTH_8U, 1);

				cvSetImageROI( pRPL->mpGray, pRPL->mInterestArea );
				cvCopy(pRPL->mpGray, ciGray);
				cvResetImageROI( pRPL->mpGray );

				cvSetImageROI( cGray, cvRect(0, 0, pRPL->mInterestArea.width,
												   pRPL->mInterestArea.height));
				cvCopy(ciGray, cGray);
				cvResetImageROI( cGray );

				cvReleaseImage(&ciGray);
			break;
		}

		cvShowImage("Input image", cGray);
		cvReleaseImage(&cGray);			
	}

	if(pRPL->mpOptFlw) {
		//  " "
		pRPL->mpOptFlw->Calculate(pRPL->mpGray);
		cvFlip(pRPL->mpScreen);
		pRPL->mpOptFlw->DrawFlow(pRPL->mpScreen, pRPL->mColoredMotion);
		cvFlip(pRPL->mpScreen);
	} else {
		//  
		pRPL->mpObjAM->SetImage(pRPL->mpGray);

		if(gpCmdArray) { pRPL->mpGateAM->SetImage(pRPL->mpGray); }

		switch(pRPL->mWorkMode) {
			case cSetTheKeyImageSize: //   
				gEOT = pRPL->mpObjAM->EstimateOpportunityForTraining(pRPL->mInterestArea);
				break;

			case cTraining: //         AVM
				pRPL->mpObjAM->Write(pRPL->mInterestArea, &pRPL->mRcgDataFile.ObjName.ObjIdx);
				break;

			case cRecognition:
				//   
				cvSetImageROI(apImage, cvRect(0,apImage->height-36,apImage->width,36));
				cvSubS(apImage, cvScalar(64,64,64), apImage);
				cvResetImageROI(apImage);

				TSearchObjectData d;
				d.RcgObj.ObjTotal = 0;
				d.RcgObj.GateHorizon = 0;
				d.RcgObj.CheckpointFound = false;
				d.OutImg    = apImage;
				d.pObjName  = &pRPL->mRcgDataFile.ObjName;
				//      
				if(pRPL->mpObjAM->GetTotalABases() > 1) {
					pRPL->mpObjAM->ObjectTracking(doSearchObject, &d);
				}

				if(gpCmdArray && pRPL->mpGateAM->GetTotalABases() > 1) {
					pRPL->mpGateAM->ObjectTracking(doSearchGate, &d, false);
					//  ?
					if(d.GateList.Total()) {// 
						int GateTotal = 0;
						TGateData* pGate;
						d.GateList.GoAtStart();
						//   
						while((pGate = d.GateList.GetNext()) != NULL) {
							if(!d.RcgObj.CheckpointFound || pGate->d.s[cPointIdx_s] == gWayIdx+1) {
								//   
								CvScalar SubColor, Color;
								double GateRate = float(ushort(pGate->d.s[gWayIdx]))/65535.0;
								double cH = 240.0 + 180.0*(1.0 - GateRate);

								SubColor = cvHLStoRGB(float(cH),0.2f,1);
								Color.val[0] = 255 - SubColor.val[0];
								Color.val[1] = 255 - SubColor.val[1];
								Color.val[3] = 255 - SubColor.val[3];

								CvRect GataRect = pGate->r;
								GataRect.y = pGate->r.y - pGate->r.height;
								cvSetImageROI(apImage, GataRect);
								cvSubS(apImage, SubColor, apImage);
								cvResetImageROI(apImage);
								cvRectangle(apImage, cvPoint(pGate->r.x,  pGate->r.y),
									        cvPoint(pGate->r.x + pGate->r.width, pGate->r.y - pGate->r.height),
									        cvScalar(0,0,255),1,8,0);

								CvPoint tP = cvPoint(pGate->r.x+8, pGate->r.y-28);
								//   
								if(pGate->d.s[cPointIdx_s]) {
									sprintf(OutStr, "Point[%d]", pGate->d.s[cPointIdx_s]);
									tP = cvPoint(pGate->r.x+(pGate->r.width>>1)-70, pGate->r.y-(pGate->r.height>>1)-12);
								} else { strcpy(OutStr, "Gate"); }
								
								cvPutText(apImage, OutStr, tP, &font, CV_RGB(255,255,0));
								cvLine(apImage, cvPoint(pGate->r.x, pGate->r.y),
								       cvPoint(pGate->r.x, GataRect.y), Color, 4,8,0);
								cvLine(apImage, cvPoint(pGate->r.x + pGate->r.width, pGate->r.y),
								       cvPoint(pGate->r.x + pGate->r.width, GataRect.y), Color, 4,8,0);

								//  ""
								double cXs = double(apImage->width>>1);
								double cXd = cXs - double(pGate->r.x+(pGate->r.width>>1));
								
								d.RcgObj.GateHorizon += (cXd/cXs)*GateRate;
								GateTotal++;
							}
						}

						if(GateTotal) { d.RcgObj.GateHorizon = d.RcgObj.GateHorizon/double(GateTotal); }

						//  ""
						if(d.RcgObj.GateHorizon) {
							int cX = apImage->width>>1;
							int hX = cX - int(cX*d.RcgObj.GateHorizon);
							CvPoint p1 = cvPoint(cX, 40), p2 = cvPoint(hX, 40);
							cvLine(apImage, p1, p2, cvScalar(0,0,0),9,8,0);
							cvLine(apImage, p1, p2, cvScalar(0,255,0),5,8,0);
							cvLine(apImage, cvPoint(cX, 35), cvPoint(cX, 45), cvScalar(0,255,0),2,8,0);
							//   
							if(gNovaGateMode) {
								d.GateList.GoAtStart();
								//    
								while((pGate = d.GateList.GetNext()) != NULL) {
									CvType192U* pData;
									CvRect iArea, ObjRect;
									iArea.x = pGate->cX - (pRPL->mGateKeySize.width>>1);
									iArea.y = pGate->cY - (pRPL->mGateKeySize.height>>1);
									iArea.width = pRPL->mGateKeySize.width;
									iArea.height = pRPL->mGateKeySize.height;

									if(iArea.x < 0 || iArea.y < 0 ||
									   iArea.x+iArea.width > apImage->width ||
									   iArea.y+iArea.height > apImage->height) { continue; }
									
									if(pRPL->mpGateAM->Read(iArea, &ObjRect, &pData)) {
										int dX = hX - ObjRect.x - (ObjRect.width>>1);
										double base = double(cX - (pRPL->mGateKeySize.width>>1));
										double k = 1.0 - (fabs(double(dX))/base);
										ushort Code = ushort(pData->s[gWayIdx]);
										double sW = k - (double(Code)/65535.0);
										uchar Hits = uchar(pData->c[cHitIdx_c+gWayIdx]);
										double kExpArg = double(Hits)/16.0;
										const double cgExpA = 0.3;
										const double cgExpB = 1.4;
										double kAge = exp(-kExpArg*cgExpA - cgExpB);
										pData->s[gWayIdx] = ushort(Code + short(sW*kAge*65535.0));

										if(Hits < 0xFF) { pData->c[cHitIdx_c+gWayIdx]++; }
									}
								}
							}
						}
					}
				}

				//   
				Navigate(pRPL->mpGray, pRPL->mpScreen, &d.RcgObj, SetCheckpoint, LogMessageFunc);
				break;
		}
        // 
		pRPL->Draw();

		//   
		static short DivCounter;
		if(!(DivCounter++ % 12)) {
			if(pRPL->mpDrawTree) { pRPL->mpDrawTree->ShowAssociativeTree(); }
		}
	}

	pRPL->LeaveProcessing();

	if(!gStopWriteDemo && gOutImg && gOutAVI) {
		//       
		cvResize(apImage, gOutImg, CV_INTER_LINEAR);
		cvFlip(gOutImg);
		cvWriteFrame(gOutAVI, gOutImg);
	}
}

//     
void CRecognitionPipeline::Comment(wxString aText)
{
	mComment = aText;
	mCommentTimer = 75;
}

//  
wxSize CRecognitionPipeline::InitCapture(int aCamIdx, TCaptureSettings aCaptSet, VidFormat* apVidFmt)
{
	//    
	int CamNum = cvcamGetCamerasCount();
	if(!CamNum) {
		//    
		wxMessageBox(_MS(_T("Could not initialize capturing...\n")
				         _T("Please, connect web-camera and then restart application.")),
				     _MST("Pipeline initializations error"), wxOK | wxICON_ERROR);
		exit(0);
	}

	mCamIdx = aCamIdx;

	if(aCaptSet == cChangeCaptureSource) { 
		int* pCamIdx = NULL;
		cvcamSelectCamera(&pCamIdx);
		if(pCamIdx) { mCamIdx = *pCamIdx; }
	}

	cvcamSetProperty(mCamIdx, CVCAM_PROP_ENABLE, CVCAMTRUE); 
	cvcamSetProperty(mCamIdx, CVCAM_PROP_RENDER, CVCAMTRUE);

	switch(aCaptSet) {
		case cChangeVideoFormat:
			cvcamGetProperty(mCamIdx, CVCAM_CAMERAPROPS, NULL);
			cvcamGetProperty(mCamIdx, CVCAM_VIDEOFORMAT, NULL);
			break;
		default:
			static VidFormat vidFmt = { 640, 480, 30.0};
			//      
			cvcamSetProperty(mCamIdx, CVCAM_PROP_SETFORMAT, (apVidFmt)?apVidFmt:&vidFmt); 
			break;
	}
	int WIDTH, HEIGHT;
	CameraDescription CamDscr;
	cvcamGetProperty(mCamIdx, CVCAM_SRCWIDTH, &WIDTH);
	cvcamGetProperty(mCamIdx, CVCAM_SRCHEIGHT, &HEIGHT);
	cvcamSetProperty(mCamIdx, CVCAM_PROP_CALLBACK, RecognitionPipelineFunc);
	cvcamGetProperty(mCamIdx, CVCAM_DESCRIPTION, &CamDscr);

	if(gpRcgPipeline) {
		wxLogMessage(wxString::Format(_MST("Capture source: \"%s\", resolution: %ux%u pixels."),
			wxString::FromAscii(CamDscr.DeviceDescription), WIDTH, HEIGHT));
		wxLogMessage(_T(""));
	}

	mImgSize = cvSize(WIDTH, HEIGHT);
	return wxSize(WIDTH, HEIGHT);
}

//   
void CRecognitionPipeline::CreatePipeline(void* apRcgData, long aObjDataOffset)
{
	mpGray = cvCreateImage( mImgSize, 8, 1 );

	mInterestArea.x = 0;
	mInterestArea.y = 0;

	ReadRcgData(apRcgData, aObjDataOffset);

	cvcamWindow mywin = (cvcamWindow)GetHandle();
	cvcamSetProperty(mCamIdx, CVCAM_PROP_WINDOW, &mywin);

	cvcamInit();
	cvcamStart();
}

#define GateSettings(pGateAM) \
	pGateAM->SetParam(ptTrackingDepth, 2);\
	pGateAM->SetParam(ptRcgMxCellBase, 19992);\
	pGateAM->SetParam(ptOptimizeLoop, 1e+9);

//   ""
void CRecognitionPipeline::InitGateMemory()
{
	mpGateAM->Create(mGateKeySize,3);
	GateSettings(mpGateAM);
}

//      
void CRecognitionPipeline::FitGateMemorySize()
{
	#define cGateAR 0.72
	CvSize GateKeySize;
	GateKeySize.width = int(mImgSize.width*cGateAR);
	GateKeySize.height = int(mImgSize.height*cGateAR);

	if(GateKeySize.width/mpGateAM->GetBaseKeySize().width != 1 ||
	   GateKeySize.height/mpGateAM->GetBaseKeySize().height != 1) {
		InitGateMemory();
	}
}


//  
template<class CvAssociativeMemory>
void LogAVM_Info(CvAssociativeMemory* apAM) {
	wxLogMessage(wxString::Format(_MST("Base key size ............. %ux%u"),   apAM->GetBaseKeySize().width,
		apAM->GetBaseKeySize().height));
	wxLogMessage(wxString::Format(_MST("Associative Levels ........ %-7u"),    apAM->GetTotalLevels()));
	wxLogMessage(wxString::Format(_MST("Associative bases ......... %-7lu"),   apAM->GetTotalABases()));
	wxLogMessage(wxString::Format(_MST("Current base index ........ %-7I64u"), apAM->GetCurIndex()));
	wxLogMessage(wxString::Format(_MST("Wr/Rd counter ............. %-7I64u"), apAM->GetWrRdCounter()));
	wxLogMessage(_T(""));
	GateSettings(apAM);
}

//   
void CRecognitionPipeline::ReadRcgData(void* apRcgData, long aObjDataOffset)
{
	mGateKeySize.width = int(mImgSize.width*cGateAR);
	mGateKeySize.height = int(mImgSize.height*cGateAR);

	//  AVM  ""
	if(!mpGateAM) { mpGateAM = new CvAssociativeMemory192U; }

	if(apRcgData && aObjDataOffset) {
		wxLogMessage(_MST("Gate recognition data:"));
		mpGateAM->ReadPackedData(apRcgData);
		LogAVM_Info(mpGateAM);
	} else { InitGateMemory(); }

	//  AVM  
	if(!mpObjAM) { mpObjAM = new CvAssociativeMemory32S; }

	if(apRcgData) {
		wxLogMessage(_MST("Object recognition data:"));
		wxLogMessage(wxString::Format(_MST("Visual objects total ...... %-7lu"), mRcgDataFile.ObjName.Total));
		mpObjAM->ReadPackedData(((char*)apRcgData)+aObjDataOffset);
		LogAVM_Info(mpObjAM);

		mKeyImageSize = mpObjAM->GetKeyImageSize();		
		mAspectRatio = double(mKeyImageSize.width)/double(mKeyImageSize.height);
	} else {
		mKeyImageSize.width  = cBaseKeyImgSize;
		mKeyImageSize.height = cBaseKeyImgSize;
		mRcgDataFile.RcgDataSize = 0;
		mpObjAM->Create(mKeyImageSize);
	}
	//   
	SetRecognitionMode();
}

//    
void CRecognitionPipeline::SetHintWin(wxString aHint, wxSize* apWinSize, bool aAtLeft) {
	int MinY;
    cvGetTextSize( mHint.ToAscii(), &mHint_Font, &mHint_TextSize, &MinY );
	mHint_TextSize.height += MinY;

	mHint_AtLeft  = aAtLeft;
	mHint         = aHint;
	mHint_WinSize = (apWinSize)?*apWinSize:
					wxSize(mHint_TextSize.width + 8, mHint_TextSize.height - cHint_OffsetY);
}

//   
wxPoint CRecognitionPipeline::DrawHint(wxCoord x, wxCoord y) {
	if(mHint != _T("")) {
		
		int static arr = 5;
		CvPoint* pt;
		CvPoint  Points[5];
		pt = Points;

		int hX = mHint_WinSize.x + cHint_OffsetX;
		int hY = mHint_WinSize.y - cHint_OffsetY;
		int _y = mpScreen->height - y;

		if(mHint_AtLeft) {
			Points[0] = cvPoint(x,  _y);			
			Points[1] = cvPoint(x-hX, _y);
			Points[2] = cvPoint(x-hX, _y-hY);
			Points[3] = cvPoint(x-cHint_OffsetX, _y-hY);
			Points[4] = cvPoint(x-cHint_OffsetX, _y-cHint_OffsetY);
			x -= (mHint_WinSize.x + cHint_OffsetX-4);
		} else {
			Points[0] = cvPoint(x,  _y);
			Points[1] = cvPoint(hX+x, _y);
			Points[2] = cvPoint(hX+x, _y-hY);
			Points[3] = cvPoint(cHint_OffsetX+x, _y-hY);
			Points[4] = cvPoint(cHint_OffsetX+x, _y-cHint_OffsetY);
			x += (cHint_OffsetX+4);
		}

		cvFillPoly( mpScreen, &pt, &arr, 1, CV_RGB(0xFF,0xFF,0xDD));
		cvPolyLine( mpScreen, &pt, &arr, 1, 1, CV_RGB(0xFF,0x80,0), 2);

		wxString OutStr;
		int StrLen = int(mHint.Length());
		for(int i=0; i<=StrLen; i++) {
			if(!(mHint[i] == '\n' || i == StrLen)) {
				OutStr += mHint[i];
			} else {
				cvPutText(mpScreen, OutStr.ToAscii(),
							cvPoint(x, mpScreen->height - y - mHint_TextSize.height), &mHint_Font, CV_RGB(0,0,0));
				y += mHint_TextSize.height; OutStr = _T("");
			}
		}
	}

	return wxPoint(x,y);
}

// 
void CRecognitionPipeline::Draw()
{
	#define cPrgBarX 120
	#define cPrgBarY 28

	int bX,bY,eX,eY,bXpb,eXpb,mXpb,bYpb,eYpb,tY,sX,sY,mX,mY;

	bX = mInterestArea.x;
	bY = mpScreen->height - mInterestArea.y;
	eX = mInterestArea.x + mInterestArea.width;
	eY = mpScreen->height - mInterestArea.y - mInterestArea.height;
	mX = mMouseCursorX;
	mY = mpScreen->height - mMouseCursorY;

	if(mMouseCursorInit && (mMouseCursorX > 0) && (mMouseCursorY > 0)) {

		switch(mWorkMode) {
			//    
			case cSetTheBaseCoordinateOfObject:
				bX = mMouseCursorX;
				bY = mpScreen->height - mMouseCursorY;
				eX = mMouseCursorX + mInterestArea.width;
				eY = mpScreen->height - mMouseCursorY - mInterestArea.height;
				cvSetImageROI(mpScreen, cvRect(bX, bY-mInterestArea.height,
										     mInterestArea.width, mInterestArea.height));
				cvConvertScale(mpScreen, mpScreen, 0.7);
				cvResetImageROI(mpScreen);
				
				cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0x00,0xFF,0xFF), 2,8,0);
				cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0x00,0x00,0xFF), 1,8,0);
				break;

			//    
			case cSetTheKeyImageSize:
				int w = mMouseCursorX - mInterestArea.x;
				int h = mMouseCursorY - mInterestArea.y;

				cvSetImageROI(mpScreen, cvRect(mX, mY, 120, 20));
				cvConvertScale(mpScreen, mpScreen, 0.7);
				cvResetImageROI(mpScreen);

				if(!mNeedKeyImageSetting) {
					int dW = w - mInterestArea.x;
					int dH = h - mInterestArea.y;
					
					if(dW > dH) {
						mInterestArea.height = mInterestArea.width/mAspectRatio;
						mInterestArea.width  = (w > mKeyImageSize.width)?w:mKeyImageSize.width;
					} else {
						mInterestArea.width = mInterestArea.height*mAspectRatio;
						mInterestArea.height = (h > mKeyImageSize.height)?h:mKeyImageSize.height;
					}
				} else {
					mInterestArea.width  = (w > mKeyImageSize.width)?w:mKeyImageSize.width;
					mInterestArea.height = (h > mKeyImageSize.height)?h:mKeyImageSize.height;

					int i;
					CvSize Key;
					Key.width  = mInterestArea.width;
					Key.height = mInterestArea.height;
					
					for(i=Key.width/10,  Key.width  = 1; i >>= 1; Key.width  <<= 1);
					for(i=Key.height/10, Key.height = 1; i >>= 1; Key.height <<= 1);

					Key.width *= 10; Key.height *= 10;

					cvPutText(mpScreen, (wxString::Format(_T("Key: %ux%u"),
			                  Key.width,Key.height)).ToAscii(), cvPoint(bX, eY-16), &mHint_Font, CV_RGB(0x00,0xFF,0x00));
				}

				cvPutText(mpScreen, (wxString::Format(_T("EOT: %f"),gEOT).ToAscii()),
					      cvPoint(mX, mY+4), &mHint_Font, CV_RGB(0xFF,0xFF,0x00));
				
				eX = mInterestArea.x + mInterestArea.width;
				eY = mpScreen->height - mInterestArea.y - mInterestArea.height;

				cvSetImageROI(mpScreen, cvRect(bX, bY-mInterestArea.height,
										     mInterestArea.width, mInterestArea.height));
				cvConvertScale(mpScreen, mpScreen, 0.7);
				cvResetImageROI(mpScreen);
				
				cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0x00,0xFF,0xFF), 2,8,0);
				cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0x00,0x00,0xFF), 1,8,0);
				break;

		}
		DrawHint(mMouseCursorX, mMouseCursorY);
	}

	bXpb = bX - ((cPrgBarX + bX - eX)>>1);
	eXpb = bXpb + cPrgBarX;
	bYpb = eY - 8;
	eYpb = eY - cPrgBarY;
	tY = bYpb - 14;
	sX = double(mInterestArea.width)*0.1;
	sY = double(mInterestArea.height)*0.1;

	switch(mWorkMode) {
		//   
		case cPrepareToTraining:
			
			mXpb = eXpb - ((mPauseLoop*cPrgBarX)/cPauseLoopMax);
			
			cvRectangle(mpScreen, cvPoint(bX+sX, bY-sY), cvPoint(eX-sX, eY+sY), CV_RGB(0x00,0x00,0xFF), 1,8,0);
			cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0xFF,0xFF,0x70), 3,8,0);
			cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0x00,0xFF,0x00), 2,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(mXpb, eYpb), CV_RGB(0xAA,0xFF,0xFF), CV_FILLED,8,0);
			cvRectangle(mpScreen, cvPoint(mXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0xB4,0xB4,0xB4), CV_FILLED,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0x00,0x00,0x00), 2,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0x00,0xFF,0x00), 1,8,0);

			cvPutText(mpScreen, "Preparing", cvPoint(bXpb+20, tY), &mHint_Font, CV_RGB(0x00,0x00,0x00));
			
			if(!mPauseLoop) {
				wxLogMessage(wxString::Format(_MS(_T("\nNow training is begun. Show the object \"%s\" ")
					                              _T("in red rectangle at the different angles.")),
							 wxString::FromAscii(mRcgDataFile.ObjName.s[mRcgDataFile.ObjName.ObjIdx])));

				mWorkMode = cTraining;
			} else { mPauseLoop--; }
			break;
			
		//   
		case cTraining:
			
			mXpb = eXpb - ((mTrainingLoop*cPrgBarX)/cTrainingLoopMax);

			cvRectangle(mpScreen, cvPoint(bX+sX, bY-sY), cvPoint(eX-sX, eY+sY), CV_RGB(0x00,0x00,0xFF), 1,8,0);
			cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0xFF,0xFF,0x70), 3,8,0);
			cvRectangle(mpScreen, cvPoint(bX, bY), cvPoint(eX, eY), CV_RGB(0xFF,0x00,0x00), 2,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(mXpb, eYpb), CV_RGB(0xFF,0xFF,0xAA), CV_FILLED,8,0);
			cvRectangle(mpScreen, cvPoint(mXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0xB4,0xB4,0xB4), CV_FILLED,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0x00,0x00,0x00), 2,8,0);
			cvRectangle(mpScreen, cvPoint(bXpb, bYpb), cvPoint(eXpb, eYpb), CV_RGB(0xFF,0x00,0x00), 1,8,0);

			cvPutText(mpScreen, "Training", cvPoint(bXpb+30, tY), &mHint_Font, CV_RGB(0x00,0x00,0x00));

			if(!mTrainingLoop) {
				SetRecognitionMode();
			} else { mTrainingLoop--; }
			break;
	}

	//     
	if(mCommentTimer) {
		cvPutText(mpScreen, mComment.ToAscii(), cvPoint(4, 4), &mComment_Font, CV_RGB(0x00,0xFF,0x00));
		mCommentTimer--;
	}
	
	Refresh();
}


//	   
void CRecognitionPipeline::OnMouseMove(wxMouseEvent& event)
{
	
	int x = event.GetPosition().x;
	int y = event.GetPosition().y;

	if(x<0 || y<0) return;
	if(mWorkMode != cSetTheKeyImageSize &&
	   (x+mInterestArea.width  > mImgSize.width ||
	    y+mInterestArea.height > mImgSize.height)) return;
	
	mMouseCursorX = x;
	mMouseCursorY = y;
	mMouseCursorInit = true;
}

//   
void CRecognitionPipeline::PrepareToTraining()
{
	SetHintWin(_T(""));
	mWorkMode = cInputObjectName;
	CDefineObjectName ObjectNameDlg(&mRcgDataFile.ObjName);
	if(ObjectNameDlg.ShowModal() == wxID_OK){
		Comment(wxString::Format(_T("Learn: %s"),
				wxString::FromAscii(mRcgDataFile.ObjName.s[mRcgDataFile.ObjName.ObjIdx])));
		wxLogMessage(wxString::Format(_MST("\nPrepare for training on object under name: \"%s\"."),
			         wxString::FromAscii(mRcgDataFile.ObjName.s[mRcgDataFile.ObjName.ObjIdx])));
		wxLogMessage(_MST("Please, expose object into green rectangle."));
		mWorkMode = cPrepareToTraining;
	} else { 
		if(mNeedKeyImageSetting || !mRcgDataFile.RcgDataSize) {
			SetTheBaseCoordinateOfObject();
		} else { SetRecognitionMode(); }
	}
}

//   
void CRecognitionPipeline::SetRecognitionMode()
{
	if(mWorkMode != cRecognition) {
		Comment(_T("Recognition is started"));
		wxLogMessage(_MST("\nRecognition process is started. Please, show an object for recognition."));
		mpObjAM->RestartTimeForOptimization();
		SetHintWin(_T(""));
		mNeedKeyImageSetting = false;
		mWorkMode = cRecognition;
	}
}

//   
void CRecognitionPipeline::SetCheckpoint(int aWayNum, ushort aGateCode) {
	CvType192U GateData, *pGateData;
	CvRect ObjRect, CentralArea;
	CentralArea.x = (mImgSize.width>>1) - (mGateKeySize.width>>1);
	CentralArea.y = (mImgSize.height>>1) - (mGateKeySize.height>>1);
	CentralArea.width = mGateKeySize.width;
	CentralArea.height = mGateKeySize.height;
	
	if(mpGateAM->GetTotalABases() > 1 && mpGateAM->Read(CentralArea, &ObjRect, &pGateData)) {
		pGateData->s[gWayIdx] = aGateCode;
		if(!pGateData->s[cPointIdx_s]) { pGateData->s[cPointIdx_s] = aWayNum; }
	} else {
		memset(&GateData, 0, sizeof(CvType192U));
		GateData.s[gWayIdx] = aGateCode;
		GateData.s[cPointIdx_s] = aWayNum;
		mpGateAM->Write(CentralArea, &GateData);
	}
}

//   "  "
void CRecognitionPipeline::OnLearnNextObj(int aWayIdx)
{
	if(gpCmdArray) {
		if(gNovaGateMode) {
			gCheckpointCounter = 512;
			Comment(wxString::Format(_T("Set the checkpoint p[%d]"),aWayIdx+1));
		} else {
			Comment(wxString::Format(_T("Walking to the point p[%d]"),aWayIdx+1));
		}
		gWayIdx = aWayIdx;
	} else { SetTheBaseCoordinateOfObject(); }

}

//  "  "
void CRecognitionPipeline::OnClearGateData()
{
	EnterProcessing();
	mpGateAM->ClearTreeData();
	LeaveProcessing();
	Comment(_T("Gate data was cleared"));
}

//   "   "
void CRecognitionPipeline::OnSetKeySize()
{
	mNeedKeyImageSetting = true;
	SetTheBaseCoordinateOfObject();
}

//   
void CRecognitionPipeline::OptimizeAssociativeTree() {
	mpObjAM->OptimizeAssociativeTree();
	mpGateAM->OptimizeAssociativeTree();
}

//     
void CRecognitionPipeline::WriteRcgData(void* apRcgData) {
	long Offset = GetObjDataOffset();

	if(Offset) { mpGateAM->WritePackedData(apRcgData); }

	mpObjAM->WritePackedData(((char*)apRcgData)+Offset);
}

//     
long CRecognitionPipeline::GetObjDataOffset() const { return (mpGateAM)?mpGateAM->GetPackedDataSize():0; }

//    
long CRecognitionPipeline::GetRcgDataSize() const {
	return mpObjAM->GetPackedDataSize() + GetObjDataOffset();
}

//   
void CRecognitionPipeline::RcgDataFileAsDefault()
{
	memset(&mRcgDataFile,0,sizeof(TRcgDataFile));
	strcpy(mRcgDataFile.Ident, cRcgDataFileIdent);
}

//    
void CRecognitionPipeline::OnMouseLeftDown(wxMouseEvent& event)
{
	wxSize HintWinSize;

	switch(mWorkMode) {
		//     
		case cSetTheBaseCoordinateOfObject:
			mInterestArea.x = event.GetPosition().x;
			mInterestArea.y = event.GetPosition().y;
			mWorkMode   = cSetTheKeyImageSize;
			HintWinSize = wxSize(120,50);
			SetHintWin(_T("Set the size\n of object"), &HintWinSize, false);

			if(mNeedKeyImageSetting) {
				wxLogMessage(_MS(_T("\nNow set the size of object that will be set as key ")
					             _T("image size for associative video memory.")));
			} else { wxLogMessage(_MST("Now set the size of object.")); }
			break;

		case cSetTheKeyImageSize:
			//     
			if(gEOT < 0.015) {
				if(wxMessageBox(wxString::Format(_MS(_T("The estimation of an opportunity for training ")
					_T("is too low (EOT: %f).\n It may cause to difficulties in recognition of this object.\n\n")
					_T("Do you want to continue?")),gEOT),_MST("Warning!"), wxYES_NO) == wxNO) {
					SetTheBaseCoordinateOfObject();
					return;
				}
			}
			//     ?
			if(mNeedKeyImageSetting) {// 
				if(wxMessageBox(wxString::Format(_MS(_T("You finished key image setting and now associative ")
					 _T("video memory will be created.\nIf you already have a trained data in file %s ")
					 _T("it will be lost.\n\nDo you want to continue?")),cAMFileName),
					 _MST("Setting of key image size for associative video memory is finished"),
					 wxYES_NO) == wxYES) {

					mNeedKeyImageSetting = false;
					mAspectRatio = double(mInterestArea.width)/double(mInterestArea.height);
					
					mKeyImageSize.width  = mInterestArea.width;
					mKeyImageSize.height = mInterestArea.height;

					mpObjAM->Create(mKeyImageSize);

					if(gpCmdArray) { gpRcgPipeline->mpObjAM->SetParam(ptTrackingDepth, 2); }

					wxLogMessage(_MST("\nAssociative memory is created."));
					wxLogMessage(wxString::Format(_MST("Base key size ............. %ux%u"),
									mpObjAM->GetBaseKeySize().width, mpObjAM->GetBaseKeySize().height));

					Comment(wxString::Format(_T("Set key size: %ux%u"), mpObjAM->GetBaseKeySize().width,
																		mpObjAM->GetBaseKeySize().height));
					PrepareToTraining();
				} else {
					if(mpObjAM->GetTotalABases() <= 1) {
						SetTheBaseCoordinateOfObject();
					} else { SetRecognitionMode(); }
				}
			} else {
				PrepareToTraining();
			}
			break;
	}
}


//=============================================================================================
//================================= Class: CRecognitionApp ====================================
//=============================================================================================

//	  
class CRecognitionApp : public wxApp
{
public:
	CMainWnd *Frame;

    virtual bool OnInit();

};

IMPLEMENT_APP(CRecognitionApp)

//  
bool CRecognitionApp::OnInit()
{
	Frame = new CMainWnd();
	Frame->Show(true);
    Frame->SetStatusText(_MST("Welcome to recognition demo application"));
    SetTopWindow(Frame);

    return true;
}

// : "  "
//=============================================================================================
//===================================== Class: CMainWnd =======================================
//=============================================================================================

//  
enum {
    wxID_RecognitionApp_Quit = wxID_EXIT,
    wxID_RecognitionApp_About = wxID_ABOUT,
    wxID_MENU_DIALOG,
	wxID_SetTheCaptureFormat,
	wxID_SetTheCaptureSource,
	wxID_Property,
	wxID_Save,
	wxID_Load,
	wxID_SetKeySize,
	wxID_LearnNextObj,
	wxID_SetWay,
	wxID_ShowInputImage,
	wxID_WriteAVIdemo,
	wxID_ShowAssociativeTree,
	wxID_NovaGateMode,
	wxID_NavigateMode,
	wxID_ClearGateData,
	wxID_ShowMotionField,
	wxID_ShowColoredMotion,
	wxID_TIMER,
    wxID_MENU_ABOUT
};

BEGIN_EVENT_TABLE(CMainWnd, wxFrame)
//	   
	EVT_MENU(wxID_SetKeySize,     CMainWnd::OnSetKeySize)
	EVT_MENU(wxID_LearnNextObj,   CMainWnd::OnLearnNextObj)
	EVT_MENU(wxID_ShowInputImage, CMainWnd::OnShowInputImage)
	EVT_MENU(wxID_WriteAVIdemo,   CMainWnd::OnWriteAVIdemo)
	EVT_MENU(wxID_ShowAssociativeTree, CMainWnd::OnShowAssociativeTree)
	EVT_MENU(wxID_NovaGateMode,      CMainWnd::OnNovaGateMode)
	EVT_MENU(wxID_NavigateMode,      CMainWnd::OnNavigateMode)
	EVT_MENU(wxID_ClearGateData,     CMainWnd::OnClearGateData)
	EVT_MENU(wxID_ShowMotionField,   CMainWnd::OnShowMotionField)
	EVT_MENU(wxID_ShowColoredMotion, CMainWnd::OnShowColoredMotion)
	EVT_MENU(wxID_SetTheCaptureFormat, CMainWnd::OnCaptureFormat)
	EVT_MENU(wxID_SetTheCaptureSource, CMainWnd::OnCaptureSource)
	EVT_MENU(wxID_Save,                CMainWnd::OnSave)
	EVT_MENU(wxID_Load,                CMainWnd::OnLoad)
    EVT_MENU(wxID_MENU_ABOUT,          CMainWnd::OnMenuAbout)
    EVT_MENU(wxID_RecognitionApp_Quit, CMainWnd::OnQuit)
    EVT_BUTTON(wxID_LearnNextObj, CMainWnd::OnLearnNextObj)
	EVT_BUTTON(wxID_SetWay, CMainWnd::OnSetWay)
	EVT_TIMER(wxID_TIMER, CMainWnd::OnTimer)
	EVT_CLOSE(CMainWnd::OnClose)
END_EVENT_TABLE()

//  
CMainWnd::CMainWnd() : wxFrame((wxFrame *)NULL, wxID_ANY, _MST("Recognition application"),
	                   wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE|wxMINIMIZE_BOX),
					   m_Timer(this, wxID_TIMER)
{
	SetIcon(wxICON(RecognitionApp));

	gpCmdArray = NULL;
	
	CreateStatusBar();
	
	//   
	wxMenu *menuMenu = new wxMenu;
    menuMenu->Append(wxID_SetKeySize, _MST("&Set key image size (New)\tCtrl-S"),
                     _MST("Set the key image size for associative video memory"));
    menuMenu->Append(wxID_LearnNextObj, _MST("&Learn object\tCtrl-L"),
                     _MST("Training of associative video memory on next object"));
	menuMenu->Append(wxID_ClearGateData, _MST("&Clear the gate data\tAlt-C"));
 	menuMenu->AppendSeparator();
	menuMenu->Append(wxID_ShowInputImage, _MST("&Show an input image\tAlt-S"), wxEmptyString, true);
    menuMenu->Append(wxID_ShowAssociativeTree, _MST("&Show associative tree\tCtrl-T"),wxEmptyString, true);
	menuMenu->Append(wxID_WriteAVIdemo, _MST("&Write AVI demo\tAlt-D"), wxEmptyString, true);
	menuMenu->Append(wxID_NovaGateMode, _MST("&Nova gate mode\tAlt-G"), wxEmptyString, true);
	menuMenu->Append(wxID_NavigateMode, _MST("&Navigate mode\tAlt-N"), wxEmptyString, true);
 	menuMenu->AppendSeparator();
	menuMenu->Append(wxID_ShowMotionField, _MST("&Show motion field\tCtrl-F"), wxEmptyString, true);
	menuMenu->Append(wxID_ShowColoredMotion, _MST("&Show colored motion\tCtrl-C"), wxEmptyString, true);
    menuMenu->AppendSeparator();
    menuMenu->Append(wxID_Save, _MST("&Save"), _MST("Save recognition data"));
    menuMenu->Append(wxID_Load, _MST("&Load"), _MST("Load recognition data"));
    menuMenu->AppendSeparator();
    menuMenu->Append(wxID_EXIT, _MST("E&xit\tAlt-X"),
		             _MST("Quit the recognition demo application"));

    wxMenu *SettingsMenu = new wxMenu;
    SettingsMenu->Append(wxID_SetTheCaptureFormat, _MST("&Capture format"),
                         _MST("Settings of capture format"));
	SettingsMenu->Append(wxID_SetTheCaptureSource, _MST("&Capture source"),
                         _MST("Settings of capture source"));
	
	wxMenu *helpMenu = new wxMenu;
    helpMenu->Append(wxID_MENU_ABOUT, _MST("&About\tF1"),
		             _MST("About recognition demo application"));

    mpMenuBar = new wxMenuBar( wxMB_DOCKABLE );
    mpMenuBar->Append(menuMenu,     _MST("&Recognition"));
	mpMenuBar->Append(SettingsMenu, _MST("&Settings"));
    mpMenuBar->Append(helpMenu,     _MST("&Help"));
	mpMenuBar->Check(wxID_ShowInputImage, false);
	mpMenuBar->Check(wxID_ShowAssociativeTree, false);
	mpMenuBar->Check(wxID_WriteAVIdemo, false);
	mpMenuBar->Check(wxID_NovaGateMode, false);
	mpMenuBar->Check(wxID_NavigateMode, false);
	mpMenuBar->Check(wxID_ShowMotionField, false);
	mpMenuBar->Check(wxID_ShowColoredMotion, false);

    SetMenuBar(mpMenuBar);

    mpPanel = new wxPanel(this, wxID_ANY);
    
    wxSizer *sizerDown = new wxStaticBoxSizer(new wxStaticBox( mpPanel, wxID_ANY,
		                                                       _MST("&Log window") ), wxVERTICAL);
	mpLBoxLog = new wxListBox(mpPanel, wxID_ANY);
	sizerDown->Add(mpLBoxLog, 1, wxGROW | wxALL, 5);
	sizerDown->SetMinSize(600, 150);

	mpRcgPlSizer = new wxBoxSizer(wxVERTICAL);
	mpRcgPlSizer->Add(gpRcgPipeline = new CRecognitionPipeline(mpPanel,0),0,wxALL,20);

	mpTopSizer = new wxBoxSizer(wxVERTICAL);
	mpTopSizer->Add(mpRcgPlSizer, 0, wxALL,0);

	mpBtnSizer = new wxBoxSizer(wxHORIZONTAL);
	mpBtnSizer->Add(mpLearnBtn = new wxButton(mpPanel,wxID_LearnNextObj,cStrLearnObj),0,wxLEFT | wxRIGHT,5);
	mpBtnSizer->Add(mpSetWayBtn = new wxButton(mpPanel,wxID_SetWay, _MST("Walking by way")),0,wxLEFT | wxRIGHT,5);

	mpWayIdx = new wxSpinCtrl(mpPanel, wxID_ANY, wxEmptyString, wxDefaultPosition, wxSize(46,wxDefaultCoord));
	mpWayIdx->SetRange(1,7);
	mpWayIdx->SetValue(1);
	mpWayIdx->Hide();
	mpSetWayBtn->Hide();

	mpBtnSizer->Add(mpWayIdx,0,wxLEFT | wxRIGHT,5);
	mpTopSizer->Add(mpBtnSizer,0,wxLEFT | wxDOWN,5);
	mpTopSizer->Add(sizerDown, 1,  wxGROW | (wxALL & ~wxTOP), 10);

	mpPanel->SetSizer(mpTopSizer);

    mpTopSizer->Fit(this);
    mpTopSizer->SetSizeHints(this);

	//     
	mpLogTarget = new LboxLogger(mpLBoxLog, wxLog::GetActiveTarget());
	wxLog::SetActiveTarget(mpLogTarget);
	
	mDataDir = wxGetHomeDir()+_T("\\AVM");
	if(!wxDirExists(mDataDir)) { wxMkdir(mDataDir); } mDataDir += _T("\\");

	long ObjDataOffset;
	void* pRcgData = Load(mDataDir+cAMFileName, &ObjDataOffset);

	gpRcgPipeline->EnterProcessing();
	gpRcgPipeline->CreatePipeline(pRcgData, ObjDataOffset);
	gpRcgPipeline->LeaveProcessing();

	if(pRcgData) { delete pRcgData; }
}

//     
void* CMainWnd::Load(wxString aFileName, long* apObjDataOffset)
{
	if(apObjDataOffset) { *apObjDataOffset = NULL; }

	static char OldDataFileIdent[] = cRcgDataFileIdent; OldDataFileIdent[6] = '_';

	char* pRcgData = NULL;
	bool RcgDataExist = false;
	wxFile* fTRcgDataFile = NULL;
	if(wxFileExists(aFileName)) {
		fTRcgDataFile = new wxFile(aFileName,wxFile::read);
		if(fTRcgDataFile->IsOpened()) {
			if(fTRcgDataFile->Read(&gpRcgPipeline->mRcgDataFile, sizeof(TRcgDataFile)) == sizeof(TRcgDataFile)) {
				if(strstr((char*)&gpRcgPipeline->mRcgDataFile.Ident,cRcgDataFileIdent) ||
				   strstr((char*)&gpRcgPipeline->mRcgDataFile.Ident,OldDataFileIdent)) {
					//   ?
					if(gpRcgPipeline->mRcgDataFile.Ident[6] == '_') {// 
						gpRcgPipeline->mRcgDataFile.Ident[6] = '-';
						TObjName ObjName = *((TObjName*)&gpRcgPipeline->mRcgDataFile.ObjDataOffset);
						gpRcgPipeline->mRcgDataFile.ObjDataOffset = 0;
						gpRcgPipeline->mRcgDataFile.ObjName = ObjName;
						fTRcgDataFile->Seek(sizeof(TRcgDataFile) - sizeof(long));
					}
					
					if(apObjDataOffset) { *apObjDataOffset = gpRcgPipeline->mRcgDataFile.ObjDataOffset; }

					if(gpRcgPipeline->mRcgDataFile.RcgDataSize) {
						pRcgData = new char[gpRcgPipeline->mRcgDataFile.RcgDataSize];
						if(fTRcgDataFile->Read(pRcgData, gpRcgPipeline->mRcgDataFile.RcgDataSize) ==
						   gpRcgPipeline->mRcgDataFile.RcgDataSize) {
							RcgDataExist = true;
						} else { if(pRcgData) delete pRcgData; pRcgData = NULL; }
					}
				}
			}
		}
	}
	if(!RcgDataExist) {
		gpRcgPipeline->RcgDataFileAsDefault();
		wxLogWarning(wxString::Format(_MST("The file %s not loaded (recognition data is absent)."),aFileName));
	}
	if(fTRcgDataFile) { delete fTRcgDataFile; }

	if(RcgDataExist && pRcgData) {
		wxLogMessage(_MST("Recognition data is loaded:"));
		wxLogMessage(wxString::Format(_MST("Load associative bases from file: %s"), aFileName));
		wxLogMessage(_T(""));
	}

	return pRcgData;
}

//     
void CMainWnd::Save(wxString aFileName)
{
	gpRcgPipeline->Comment(_T("Save recognition data"));
	gpRcgPipeline->OptimizeAssociativeTree();
	gpRcgPipeline->mRcgDataFile.RcgDataSize = gpRcgPipeline->GetRcgDataSize();
	gpRcgPipeline->mRcgDataFile.ObjDataOffset = gpRcgPipeline->GetObjDataOffset();

	if(gpRcgPipeline->mRcgDataFile.RcgDataSize > 40) {

		wxFile* fTRcgDataFile = new wxFile(aFileName, wxFile::write);
		fTRcgDataFile->Seek(0);

		if (fTRcgDataFile->IsOpened()) {
			fTRcgDataFile->Write(&gpRcgPipeline->mRcgDataFile,sizeof(TRcgDataFile));
			char* p = new char[gpRcgPipeline->mRcgDataFile.RcgDataSize];
			gpRcgPipeline->WriteRcgData(p);
			fTRcgDataFile->Write(p,gpRcgPipeline->mRcgDataFile.RcgDataSize);
			delete p;
		}

		fTRcgDataFile->Close();
		delete fTRcgDataFile;
	} else {
		wxMessageBox(_MST("There is no recognition data for saving"),
					 _MST("Save error"), wxOK | wxICON_ERROR);
	}
}

//   - ""
void CMainWnd::OnQuit(wxCommandEvent& WXUNUSED(event))
{
	wxCloseEvent event;
	OnClose(event);
}

//   
void CMainWnd::OnClose(wxCloseEvent& WXUNUSED(event))
{
	if(gOutAVI) { gStopWriteDemo = true; m_Timer.Start(2000, true); }

	int Cmd = wxMessageBox(wxString::Format(_MS(_T("Do you want to save recognition ")
		                   _T("data to file %s before exit?")),cAMFileName),
						   _MST("Program finishing"),wxYES_NO | wxCANCEL);
	wxTimerEvent Event;
	OnTimer(Event);

	if(Cmd == wxYES || Cmd == wxNO) {
		if(Cmd == wxYES) {
			gpRcgPipeline->EnterProcessing();
			Save(mDataDir+cAMFileName);
			gpRcgPipeline->LeaveProcessing();
		}
		
		delete gpRcgPipeline;
		Destroy();

		FinishNavigateMode();

		exit(0);
	}
}

//  "  "
void CMainWnd::OnLearnNextObj(wxCommandEvent& WXUNUSED(event))
{
	if(!gpCmdArray) {
		wxLogMessage(_MST("\nTraining of associative video memory on next object."));
		if(wxMessageBox(wxString::Format(_MS(_T("Do you want to save recognition ")
			_T("data to file %s before training?")),cAMFileName),
			_MST("Preparing to training"),wxYES_NO) == wxYES) {
			Save(mDataDir+cAMFileName);
		}
	}

	gpRcgPipeline->OnLearnNextObj(mpWayIdx->GetValue()-1);
}

//  " "
void CMainWnd::OnSetWay(wxCommandEvent& WXUNUSED(event))
{
	gWayIdx = mpWayIdx->GetValue()-1;
	gpRcgPipeline->Comment(wxString::Format(_T("Walking by way[%d]"),gWayIdx+1));
}

//  "  "
void CMainWnd::OnShowInputImage(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		gShowInputImage = true;
		cvNamedWindow("Input image", 1);
	} else {
		gShowInputImage = false;
		cvDestroyWindow("Input image");
	}
}

//  "  "
void CMainWnd::OnShowAssociativeTree(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		if(gpRcgPipeline->mpDrawTree) { delete gpRcgPipeline->mpDrawTree; }
		gpRcgPipeline->mpDrawTree = new CvDrawAssociativeTree(gpRcgPipeline->mpObjAM);
	} else {
		if(gpRcgPipeline->mpDrawTree) {
			delete gpRcgPipeline->mpDrawTree;
			gpRcgPipeline->mpDrawTree = NULL;
		}
	}	
}

//   
void CMainWnd::StartNavigateMode(bool aNovaGateMode)
{
	CvSize iSize = gpRcgPipeline->GetImgSize();

	if(!(iSize.width == 320 && iSize.height == 240) &&
		wxMessageBox(_MS(_T("For this mode would be better to change resolution.\n\n")
					 _T("Do you want to set resolution 320x240 pixels?\n")),
					 _MST("Video input resolution changing"),wxYES_NO) == wxYES) {

		static VidFormat NavigateVidFmt = { 320, 240, 25.0};
		ChangeCapture(cDefault, &NavigateVidFmt);
	}

	gpRcgPipeline->FitGateMemorySize();

	//     
	bool AlreadyRunning;
	HANDLE hMutexOneInstance = ::CreateMutex( NULL, FALSE,
		_T("Robot_Controller_V1-073FA540-B11D-12D3-BC37-009097709574"));
	AlreadyRunning = (::GetLastError() == ERROR_ALREADY_EXISTS || 
						  ::GetLastError() == ERROR_ACCESS_DENIED);
	::CloseHandle(hMutexOneInstance);
	
	if (!AlreadyRunning) {
		if(WinExec("RobotController.exe",SW_SHOW) < 31) {
			wxMessageBox(_MST("The robot control driver (RobotController.exe) is not loaded."),
				_MST("Robot control driver"), wxOK | wxICON_ERROR);
			mpMenuBar->Check(wxID_NovaGateMode, false);
			mpMenuBar->Check(wxID_NavigateMode, false);
			return;			
		}
	}
	
	::StartNavigateMode(aNovaGateMode);

	if(aNovaGateMode) { mpSetWayBtn->Show(); } else { mpSetWayBtn->Hide(); }

	mpWayIdx->Show();
	mpBtnSizer->Layout();
}

//   
void CMainWnd::FinishNavigateMode()
{
	::FinishNavigateMode();

	mpLearnBtn->SetLabel(cStrLearnObj);
	mpSetWayBtn->Hide();
	mpWayIdx->Hide();
	mpBtnSizer->Layout();
}

//  "  "
void CMainWnd::OnNovaGateMode(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		mpMenuBar->Check(wxID_NavigateMode, false);
		gpRcgPipeline->SetRecognitionMode();
		mpLearnBtn->SetLabel(_MST("Set checkpoint"));
		StartNavigateMode(true);
		gpRcgPipeline->Comment(_T("Nova gate mode on"));
	} else {
		FinishNavigateMode();
		gpRcgPipeline->OptimizeAssociativeTree();
		gpRcgPipeline->Comment(_T("Nova gate mode off"));
	}
}

//  " "
void CMainWnd::OnNavigateMode(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		mpMenuBar->Check(wxID_NovaGateMode, false);
		mpLearnBtn->SetLabel(_MST("Walking to the point "));
		StartNavigateMode();
		gpRcgPipeline->mpObjAM->SetParam(ptTrackingDepth, 2);
		gpRcgPipeline->Comment(_T("Navigate mode on"));
	} else {
		FinishNavigateMode();
		gpRcgPipeline->mpObjAM->SetParam(ptTrackingDepth, 40);
		gpRcgPipeline->Comment(_T("Navigate mode off"));
	}	
}

//  "  "
void CMainWnd::OnClearGateData(wxCommandEvent& WXUNUSED(event))
{
	if(wxMessageBox(_MST("Do you want to delete gate recognition data?"),
		_MST("Gate data clearing"),wxYES_NO) == wxYES) {
		gpRcgPipeline->OnClearGateData();			
	}
}

//  "  "
void CMainWnd::OnShowMotionField(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		if(!gpRcgPipeline->mpOptFlw) {
			if(gpCmdArray) {
				mpMenuBar->Check(wxID_ShowMotionField, false);
				return;
			}
			gpRcgPipeline->mpOptFlw = new CvOpticalFlow(gpRcgPipeline->GetImgSize());
		}
		gpRcgPipeline->mColoredMotion = false;
	} else {
		gpRcgPipeline->EnterProcessing();
		if(gpRcgPipeline->mpOptFlw) {
			delete gpRcgPipeline->mpOptFlw;
			gpRcgPipeline->mpOptFlw = NULL;
		}
		gpRcgPipeline->LeaveProcessing();
	}
	mpMenuBar->Check(wxID_ShowColoredMotion, false);
}

//  "   "
void CMainWnd::OnShowColoredMotion(wxCommandEvent& event)
{
	if(event.IsChecked()) {
		if(!gpRcgPipeline->mpOptFlw) {
			if(gpCmdArray) {
				mpMenuBar->Check(wxID_ShowColoredMotion, false);
				return;
			}
			gpRcgPipeline->mpOptFlw = new CvOpticalFlow(gpRcgPipeline->GetImgSize());
		}
		gpRcgPipeline->mColoredMotion = true;
	} else {
		gpRcgPipeline->EnterProcessing();
		if(gpRcgPipeline->mpOptFlw) {
			delete gpRcgPipeline->mpOptFlw;
			gpRcgPipeline->mpOptFlw = NULL;
		}
		gpRcgPipeline->LeaveProcessing();
	}
	mpMenuBar->Check(wxID_ShowMotionField, false);
}

//  "  "
void CMainWnd::OnWriteAVIdemo(wxCommandEvent& event)
{
	wxString FileName = mDataDir + _T("avm_demo.avi");
	if(event.IsChecked() && gStopWriteDemo) {
		CvSize plImgSize = gpRcgPipeline->GetImgSize();
		CvSize ImgSize;
		ImgSize.width  = 320;
		ImgSize.height = (ImgSize.width*plImgSize.height)/plImgSize.width;
		
		wxRemoveFile(FileName);

		gOutAVI = cvCreateVideoWriter(FileName.ToAscii(), CV_FOURCC('x','v','i','d'), 25.0, ImgSize, 1 );
		if(!gOutAVI) {
			gOutAVI = cvCreateVideoWriter(FileName.ToAscii(), -1, 25.0, ImgSize, 1 );
		}
		gOutImg = cvCreateImage( ImgSize, 8, 3 );
		
		gpRcgPipeline->Comment(wxString::Format(_T("Resolution: %ux%u"),
									plImgSize.width, plImgSize.height));
		gStopWriteDemo = false;
		wxLogMessage(wxString::Format(_MST("\nDemo writing process is started (file: %s)."),FileName));
	} else {
		gStopWriteDemo = true;
		m_Timer.Start(3000, true);
	}
}

//  
void CMainWnd::OnTimer(wxTimerEvent& WXUNUSED(event))
{
	if(gOutAVI) {
		cvReleaseVideoWriter(&gOutAVI);
		cvReleaseImage(&gOutImg);
		wxLogMessage(_MST("\nDemo writing process is finished."));
	}
}

//  "   "
void CMainWnd::OnSetKeySize(wxCommandEvent& WXUNUSED(event)) { gpRcgPipeline->OnSetKeySize(); }

//    
void CMainWnd::ChangeCapture(TCaptureSettings aCaptSet, VidFormat* apVidFmt)
{
	gpRcgPipeline->EnterProcessing();
	//     
	void* pRcgData = NULL;
	TRcgDataFile RcgDataFile = gpRcgPipeline->mRcgDataFile;

	int CamIdx = gpRcgPipeline->GetCameraIndex();
	
	if(gpRcgPipeline->GetRcgDataSize() > 40) {
		RcgDataFile.RcgDataSize = gpRcgPipeline->GetRcgDataSize();
		RcgDataFile.ObjDataOffset = gpRcgPipeline->GetObjDataOffset();
		pRcgData = new char[gpRcgPipeline->GetRcgDataSize()];
		gpRcgPipeline->WriteRcgData(pRcgData);
	}
	//   
	mpRcgPlSizer->Remove(gpRcgPipeline);
	gpRcgPipeline->LeaveProcessing();
	delete gpRcgPipeline;

	wxLogMessage(_T(""));
	//   
	mpRcgPlSizer->Add(gpRcgPipeline = new CRecognitionPipeline(mpPanel, CamIdx, aCaptSet, apVidFmt),0,wxALL,20);
	
	wxBusyInfo info(_MST("Working, please wait..."), this);

	mpRcgPlSizer->Layout();
	mpTopSizer->Layout();

	gpRcgPipeline->mRcgDataFile = RcgDataFile;
	gpRcgPipeline->CreatePipeline(pRcgData, RcgDataFile.ObjDataOffset);

	if(pRcgData) { delete pRcgData; }
}

//  "   "
void CMainWnd::OnCaptureFormat(wxCommandEvent& WXUNUSED(event)) { ChangeCapture(cChangeVideoFormat); }
//  "   "
void CMainWnd::OnCaptureSource(wxCommandEvent& WXUNUSED(event)) { ChangeCapture(cChangeCaptureSource); }

//  "  "
void CMainWnd::OnSave(wxCommandEvent& WXUNUSED(event))
{
	gpRcgPipeline->EnterProcessing();
    wxFileDialog SaveDlg(this,
                        _MST("Save recognition data file"),
                        wxEmptyString,
                        cAMFileName,
                        _T("AVM data files (*.dat)|*.dat"),
                        wxSAVE|wxOVERWRITE_PROMPT);

    SaveDlg.SetFilterIndex(1);

    if (SaveDlg.ShowModal() == wxID_OK) {
 		Save(SaveDlg.GetPath());
		wxLogMessage(_MST("File: \"%s\" was saved."), SaveDlg.GetPath());
    }
	gpRcgPipeline->LeaveProcessing();
}

//  "  "
void CMainWnd::OnLoad(wxCommandEvent& WXUNUSED(event))
{
	gpRcgPipeline->EnterProcessing();

	wxFileDialog LoadDlg(this, _MST("Load recognition data file"),
                  wxEmptyString, wxEmptyString, _T("AVM data files (*.dat)|*.dat"));

	if (LoadDlg.ShowModal() == wxID_OK) {

		LoadDlg.SetDirectory(mDataDir);
		LoadDlg.CentreOnParent();

		wxLogMessage(_T(""));
		long ObjDataOffset;
		void* pRcgData = Load(LoadDlg.GetPath(), &ObjDataOffset);
		if(pRcgData) {
			gpRcgPipeline->ReadRcgData(pRcgData, ObjDataOffset);
		} else {
			gpRcgPipeline->RcgDataFileAsDefault();
			wxMessageBox(wxString::Format(_MST("Could not load file: %s, file corrupted."),
						  LoadDlg.GetFilename()), _MST("Load error"), wxOK | wxICON_ERROR);
		}

		if(pRcgData) { delete pRcgData; }
	}
	gpRcgPipeline->LeaveProcessing();
}

//  "    "
void CMainWnd::OnMenuAbout(wxCommandEvent& WXUNUSED(event))
{
	wxStandardPathsBase& stp = wxStandardPaths::Get();
	wxSetWorkingDirectory(stp.GetDataDir());
	ShellExecute(HWND(GetHandle()), _T("open"), stp.GetDataDir() + _T("\\about.chm"),NULL,NULL,SW_SHOW);
}





