//     ======================== RecognitionApp.h ============================
//     **********************************************************************
//     Author: Yeremeyev Dmitriy Vladimirovich *** mailto:edvmail@mail.ru
//     Last update : 04.01.2010
//     **********************************************************************


#if !defined(_RecognitionApp_H_INCLUDED_)
#define _RecognitionApp_H_INCLUDED_

#include "wx/wx.h"
#include <wx/file.h>
#include "AssociativeMemory192u.h"
#include "AssociativeMemory32s.h"
#include "ListInterface.h"
#include "AVM_drawing.h"
#include "OpticalFlow.h"
#include "Navigator.h"
#include "cvcam.h"

// : " "
//=============================================================================================
//============================= Class: CRecognitionPipeline ===================================
//=============================================================================================

#define cRcgDataFileIdentStrLen 12             //      
#define cRcgDataFileIdent       "RcgApp-v0.5"  //     
#define cAMFileName             _T("avm.dat")  //    
#define cObjMaxNum              1000           //    
#define cObjNameLenMax          64             //      
#define cBaseKeyImgSize         80             //    
#define cPointIdx_s             11             //     
#define cHitIdx_c               14             //     

//    
struct TObjName {
	char s[cObjMaxNum][cObjNameLenMax]; //   
	long ObjIdx;                        //    
	long Total;                         //     
};

//    
struct TRcgDataFile {
	char     Ident[cRcgDataFileIdentStrLen]; //  
	long     RcgDataSize;                    //    ( )
	long     ObjDataOffset;                  //    
	TObjName ObjName;                        //   
};

//   
typedef enum {
	cSetTheBaseCoordinateOfObject = 0, //  ()   
	cInputObjectName,                  //   
	cSetTheKeyImageSize,               //    
	cPrepareToTraining,                //   
	cTraining,                         //  ""
	cRecognition                       //  ""
} TWorkMode;

//    
typedef enum {
	cDefault = 0,         //   
	cChangeVideoFormat,   //    
	cChangeCaptureSource  //    
} TCaptureSettings;


// : " "
class CRecognitionPipeline : public wxStaticText {
public:
	TRcgDataFile             mRcgDataFile;   //   
	CvDrawAssociativeTree*   mpDrawTree;     //      
	bool                     mColoredMotion; //   "  "
	CvOpticalFlow*           mpOptFlw;       //    " "
	CvAssociativeMemory32S*  mpObjAM;        //   AVM ,   
	CvAssociativeMemory192U* mpGateAM;       //   AVM ,   ""
	CvSize                   mGateKeySize;   //   ""

private:	
	CvSize    mImgSize;     //   	
	TWorkMode mWorkMode;    //  
	IplImage* mpScreen;     //    
	IplImage* mpGray;       //    - 
	CvRect    mInterestArea;//   (  )
	long      mPauseLoop;   //    
	long      mTrainingLoop;//   
	bool      mNeedKeyImageSetting;         //      
	wxString  mHint;                        //   ( )
	bool      mMouseCursorInit;             //    
	wxCoord   mMouseCursorX, mMouseCursorY; //   
	wxSize    mHint_WinSize;                //   

	#define   cHint_OffsetX 20 //     X
	#define   cHint_OffsetY 10 //     X
	CvFont    mHint_Font;      //   
	CvSize    mHint_TextSize;  //   
	bool      mHint_AtLeft;    //      
	double    mAspectRatio;    //     
	CvSize    mKeyImageSize;   //   
	int       mCamIdx;         //  
	CvFont    mComment_Font;   //   
	wxString  mComment;        //  
	int       mCommentTimer;   //     

	wxCriticalSection mProcessing; //     

public:
	//  
	CRecognitionPipeline(wxWindow* parent, int aCamIdx,
						 TCaptureSettings aCaptSet = cDefault, VidFormat* apVidFmt = NULL);
	//  
	~CRecognitionPipeline();
	//  
	wxSize InitCapture(int aCamIdx, TCaptureSettings aCaptSet, VidFormat* apVidFmt);
	//   
	void CreatePipeline(void* apRcgData = NULL, long aObjDataOffset = 0);
	//   ""
	void InitGateMemory();
	//      
	void FitGateMemorySize();
	//   
	void ReadRcgData(void* apRcgData, long aObjDataOffset = 0);
	//    
	void SetTheBaseCoordinateOfObject();
	//   
	void PrepareToTraining();
	//   
	void SetRecognitionMode();
	//   
	void SetCheckpoint(int aWayNum, ushort aGateCode);
	//   "  "
	void OnLearnNextObj(int aWayIdx);
	//  "  "
	void OnClearGateData();
	//   "   "
	void OnSetKeySize();
	//   
	void OptimizeAssociativeTree();
	//     
	void WriteRcgData(void* apRcgData);
	//    
	long GetRcgDataSize() const;
	//     
	long GetObjDataOffset() const;
	//   
	void RcgDataFileAsDefault();
	//     
	void Comment(wxString aText);
	// 
	void Draw();
	//    
	void OnMouseMove(wxMouseEvent& event);
	void OnMouseLeftDown(wxMouseEvent& event);
	//    
	void SetHintWin(wxString aTitle, wxSize* apWinSize = NULL, bool aAtLeft = true);
	//   
	wxPoint DrawHint(wxCoord x, wxCoord y);
	//       
	void EnterProcessing() { mProcessing.Enter(); }
	//       
	void LeaveProcessing() { mProcessing.Leave(); }
	//   
	int GetCameraIndex() const { return mCamIdx; }
	//    
	CvSize GetImgSize() const { return mImgSize; }
	// -  
	friend void RecognitionPipelineFunc(IplImage* image);

private:
    DECLARE_EVENT_TABLE()
};


//=============================================================================================
//===================================== Class: CMainWnd =======================================
//=============================================================================================

// : "  "
class CMainWnd : public wxFrame
{
	wxString    mDataDir;     //      
    wxPanel*    mpPanel;      //  " "
    wxListBox*  mpLBoxLog;    //    
    wxLog*      mpLogTarget;  //     
	wxBoxSizer* mpRcgPlSizer; //      
	wxBoxSizer* mpTopSizer;   //     
	wxBoxSizer* mpBtnSizer;   //     
	wxSpinCtrl* mpWayIdx;     //  
	wxMenuBar*  mpMenuBar;    //     
	wxButton*   mpLearnBtn;   //    " "
	wxButton*   mpSetWayBtn;  //    " "

	//   
	void StartNavigateMode(bool aNovaGateMode = false);
	//   
	void FinishNavigateMode();

public:
	//  
    CMainWnd();

    //     
	void* Load(wxString aFileName, long* apObjDataOffset = NULL);
	//     
	void  Save(wxString aFileName);
	//    
	void ChangeCapture(TCaptureSettings aCaptSet, VidFormat* apVidFmt = NULL);
	//   "  "
    void OnQuit(wxCommandEvent& event);
    void OnClose(wxCloseEvent& event);
    //  "  "
    void OnLearnNextObj(wxCommandEvent& event);
    //  " "
    void OnSetWay(wxCommandEvent& event);
	//  "   "
    void OnSetKeySize(wxCommandEvent& event);
	//  "  "
	void OnShowInputImage(wxCommandEvent& event);
	//  "  "
	void OnWriteAVIdemo(wxCommandEvent& event);
	//  "  "
	void OnShowAssociativeTree(wxCommandEvent& event);
	//  "  "
	void OnNovaGateMode(wxCommandEvent& event);
	//  " "
	void OnNavigateMode(wxCommandEvent& event);
	//  "  "
	void OnClearGateData(wxCommandEvent& event);
	//  "  "
	void OnShowMotionField(wxCommandEvent& event);
	//  "   "
	void OnShowColoredMotion(wxCommandEvent& event);
	//  "   "
	void OnCaptureFormat(wxCommandEvent& event);
	//  "   "
	void OnCaptureSource(wxCommandEvent& event);
	//  "  "
	void OnSave(wxCommandEvent& event);
	//  "  "
	void OnLoad(wxCommandEvent& event);
	//  "    "
    void OnMenuAbout(wxCommandEvent& event);
	//  
	void OnTimer(wxTimerEvent& event);

private:
	wxTimer m_Timer;

    DECLARE_EVENT_TABLE()
};

//=============================================================================================
//==================================== Class: LboxLogger ======================================
//=============================================================================================

//   
class LboxLogger : public wxLog
{
public:
    LboxLogger(wxListBox *lbox, wxLog *logOld)
    {
        m_lbox = lbox;
        m_logOld = logOld;
    }

    virtual ~LboxLogger()
    {
        wxLog::SetActiveTarget(m_logOld);
    }

private:
    //   
    virtual void DoLog(wxLogLevel level, const wxChar *szString, time_t t)
    {
        if ( level == wxLOG_Trace ) {
            if ( m_logOld ) { ((LboxLogger *)m_logOld)->DoLog(level, szString, t); }
        } else {
            wxLog::DoLog(level, szString, t);
        }
    }
	//   
    virtual void DoLogString(const wxChar *szString, time_t WXUNUSED(t))
    {
		
#ifdef _UNICODE
		int StrLen = wcslen(szString);
#else
		int StrLen = strlen(szString);
#endif		
		wxString msg = _T("");
		for(int i=0; i<StrLen; i++) {
			if(szString[i] == '\n') {
				m_lbox->Append(msg);
				msg = _T("");
			} else { msg += szString[i]; }
		}

		m_lbox->Append(msg);
		m_lbox->SetFirstItem(m_lbox->GetCount()-1);
    }

    //    
    wxListBox *m_lbox;

    //     
    wxLog *m_logOld;
};

#endif // !defined(_RecognitionApp_H_INCLUDED_)
